-------------------------------------------------------------------------------
-- DOS.E    : Standard DOS directory and file maipulation functions          --
-------------------------------------------------------------------------------
-- Author   : Daniel Berstein <daber@pair.com>                               --
-- Date     : June 8, 1998                                                   --
-- Status   : Freeware                                                       --
-- Thanks   : Jacques Deschenes, some portions copied from doswrap.e         --
-------------------------------------------------------------------------------
-- Changes  :                                                                --
-- June 8, 1998 - Added long filename support. Added shortpath(), deltree()  --
-- June 6, 1998 - First release                                              --
-------------------------------------------------------------------------------
-- Functions:                                                                --
--      All functions return on error 0 (false)                              --
--      On success all functions except mktemp() & shortpath() return 1      --
--      (true)                                                               --
-------------------------------------------------------------------------------
--      i = chdir(dirname)                                                   --
--          Changes the current directory                                    --
-------------------------------------------------------------------------------
--      i = chmod(file,attributes)                                           --
--          Changes file attributes                                          --
--          Posbible values for attrib:                                      --
--              0    Normal file                                             --
--              1    Read only file                                          --
--              2    Hidden file                                             --
--              4    Sistem file                                             --
--              8    Volume label                                            --
--              16   Directory                                               --
--              32   Backup file                                             --
-------------------------------------------------------------------------------
--      i = copy(source,dest)                                                --
--          Copies src to dest                                               --
-------------------------------------------------------------------------------
--      i = delete(filepath)                                                 --
--          Deletes a file                                                   --
-------------------------------------------------------------------------------
--      i = deltree(dirname)                                                 --
--          Deletes a directory and all it's files/subdirectories            --
--          NOTE: Can't delete root or current directory                     --
-------------------------------------------------------------------------------
--      i = dosver()                                                         --
--          Returns DOS version number                                       --
-------------------------------------------------------------------------------
--      i = mkdir(dirname)                                                   --
--          Creates a directory                                              --
-------------------------------------------------------------------------------
--      x = mktemp(path)                                                     --
--          Creates a temporary (unique) file at the given path. If no path  --
--          is provided (path = ""), the file is created on the current      --
--          directory. On success:                                           --
--          x = Filename of created file                                     --
--          NOTE: Created file must be opened before use                     --
--          Possible values for attribute:                                   --
--              0       Normal file                                          --
--              1       Read only file                                       --
--              2       Hidden file                                          --
--              4       System file                                          --
--              32      Backup file                                          --
-------------------------------------------------------------------------------
--      i = move(old,new)                                                    --
--          Moves file old to new. Can be used as rename among different     --
--          disk drives.                                                     --
-------------------------------------------------------------------------------
--      i = rename(old,new)                                                  --
--          Renames file old to new.                                         --
--          NOTE: Can't rename among different disk drives                   --
-------------------------------------------------------------------------------
--      i = rmdir(dirname)                                                   --
--          Removes a directory                                              --
--          NOTE: Only removes empty directories. Can't remove current       --
--          directory. See also deltree                                      --
-------------------------------------------------------------------------------
--      x = shortpath(path)                                                  --
--          Returns the short path given a long filename path                --
--          NOTE: Under DOS versions < 7.0 the same path is returned         --
-------------------------------------------------------------------------------
-------------------------------------------------------------------------------

include machine.e
include file.e

constant true = 1
constant false = 0

type string (sequence s)
    for loop = 1 to length(s) do
	if sequence(s[loop]) then
	    return false
	end if
    end for
    return true
end type

global function dosver()
    sequence reg_list
    reg_list = repeat(0,10)
    reg_list[REG_AX] = #3000
    reg_list = dos_interrupt(#21,reg_list)
    return remainder(reg_list[REG_AX],256) + floor(reg_list[REG_AX]/256)/100
end function

global function shortpath(string path)
    atom low_buff_long, low_buff_short, i
    sequence reg_list
    if dosver() < 7 then
	return path
    end if
    low_buff_long = allocate_low(length(path) + 1)
    if not low_buff_long then
	return false
    end if
    low_buff_short = allocate_low(129)
    if not low_buff_short then
	free_low(low_buff_long)
	return false
    end if
    poke(low_buff_long, path & 0)
    reg_list = repeat(0,10)
    reg_list[REG_AX] = #7160
    reg_list[REG_CX] = #0001
    reg_list[REG_DS] = floor(low_buff_long / 16)
    reg_list[REG_SI] = remainder(low_buff_long, 16)
    reg_list[REG_ES] = floor(low_buff_short / 16)
    reg_list[REG_DI] = remainder(low_buff_short, 16)
    reg_list[REG_FLAGS] = or_bits(reg_list[REG_FLAGS], 1)
    reg_list = dos_interrupt(#21, reg_list)
    free_low(low_buff_long)
    if and_bits(reg_list[REG_FLAGS], 1) != 0 then
	return false
    else
	i = low_buff_short
	path = ""
	while peek(i) != 0 do
	    path = path & peek(i)
	    i = i + 1
	end while
	free_low(low_buff_short)
	return path
    end if
end function

global function mkdir(string dirname)
    atom low_buff
    sequence reg_list
    low_buff = allocate_low(length (dirname) + 1)
    if not low_buff then
	return false
    end if
    poke(low_buff, dirname & 0)
    reg_list = repeat(0,10)
    if dosver() < 7 then
	reg_list[REG_AX] = #3900
    else
	reg_list[REG_AX] = #7139
    end if
    reg_list[REG_DS] = floor(low_buff / 16)
    reg_list[REG_DX] = remainder(low_buff, 16)
    reg_list[REG_FLAGS] = or_bits(reg_list[REG_FLAGS], 1)
    reg_list = dos_interrupt(#21, reg_list)
    free_low(low_buff)
    if and_bits(reg_list[REG_FLAGS], 1) != 0 then
	return false
    else
	return true
    end if
end function

global function chdir(string dirname)
    atom low_buff
    sequence reg_list
    low_buff = allocate_low(length(dirname) + 1)
    if not low_buff then
	return false
    end if
    poke(low_buff, dirname & 0)
    reg_list = repeat(0,10)
    if dosver() < 7 then
	reg_list[REG_AX] = #3B00
    else
	reg_list[REG_AX] = #713B
    end if
    reg_list[REG_DS] = floor(low_buff / 16)
    reg_list[REG_DX] = remainder(low_buff, 16)
    reg_list[REG_FLAGS] = or_bits(reg_list[REG_FLAGS], 1)
    reg_list = dos_interrupt(#21, reg_list)
    free_low(low_buff)
    if and_bits(reg_list[REG_FLAGS], 1) != 0 then
	return false
    else
	return true
    end if
end function

global function rmdir(string dirname)
    atom low_buff
    sequence reg_list
    low_buff = allocate_low(length(dirname) + 1)
    if not low_buff then
	return false
    end if
    poke(low_buff, dirname & 0)
    reg_list = repeat(0,10)
    if dosver() < 7 then
	reg_list[REG_AX] = #3A00
    else
	reg_list[REG_AX] = #713A
    end if
    reg_list[REG_DS] = floor(low_buff / 16)
    reg_list[REG_DX] = remainder(low_buff, 16)
    reg_list[REG_FLAGS] = or_bits(reg_list[REG_FLAGS], 1)
    reg_list = dos_interrupt(#21, reg_list)
    free_low(low_buff)
    if and_bits(reg_list[REG_FLAGS], 1) != 0 then
	return false
    else
	return true
    end if
end function

global function mktemp(string path, integer attrib)
    atom low_buff
    object i
    sequence reg_list
    if length(path) = 0 then
	path = current_dir()
    end if
    i = shortpath(path)
    if atom(i) then
	return false
    end if
    path = i
    if path[length(path)] != '\\' then
	path = path & '\\'
    end if
    low_buff = allocate_low(length(path) + 13 + 1)
    if not low_buff then
	return false
    end if
    poke(low_buff, path & 0)
    reg_list = repeat(0,10)
    reg_list[REG_AX] = #5A00
    reg_list[REG_CX] = attrib
    reg_list[REG_DS] = floor(low_buff / 16)
    reg_list[REG_DX] = remainder(low_buff, 16)
    reg_list[REG_FLAGS] = or_bits(reg_list[REG_FLAGS], 1)
    reg_list = dos_interrupt(#21, reg_list)
    i = low_buff
    free_low(low_buff)
    path = ""
    while peek(i) != 0 do
	path = path & peek(i)
	i = i + 1
    end while
    if and_bits(reg_list[REG_FLAGS], 1) != 0 then
	return false
    else
	reg_list[REG_BX] = reg_list[REG_AX]
	reg_list[REG_AX] = #3E00
	reg_list[REG_FLAGS] = or_bits(reg_list[REG_FLAGS], 1)
	reg_list = dos_interrupt(#21, reg_list)
	return path
    end if
end function

global function delete(string filename)
    atom low_buff
    sequence reg_list
    low_buff = allocate_low(length(filename) + 1)
    if not low_buff then
	return false
    end if
    poke(low_buff, filename & 0)
    reg_list = repeat(0,10)
    if dosver() < 7 then
	reg_list[REG_AX] = #4100
    else
	reg_list[REG_AX] = #7141
    end if
    reg_list[REG_DS] = floor(low_buff / 16)
    reg_list[REG_DX] = remainder(low_buff, 16)
    reg_list[REG_SI] = #0000
    reg_list[REG_FLAGS] = or_bits(reg_list[REG_FLAGS], 1)
    reg_list = dos_interrupt(#21, reg_list)
    free_low(low_buff)
    if and_bits(reg_list[REG_FLAGS], 1) != 0 then
	return false
    else
	return true
    end if
end function

global function copy(string src, string dest)
    integer fn1, fn2
    object i
    i = shortpath(src)
    if atom(i) then
	return false
    end if
    src = i
    i = shortpath(dest)
    if atom(i) then
	return false
    end if
    dest = i
    fn1 = open(src, "rb")
    if fn1 = -1 then
	return false
    end if
    fn2 = open(dest, "wb")
    if fn2 = -1 then
	close(fn1)
	return false
    end if
    i = getc(fn1)
    while i != -1 do
	puts(fn2,i)
	i = getc(fn1)
    end while
    close(fn2)
    close(fn1)
end function

global function rename(string oldname, string newname)
    atom low_buff_old, low_buff_new
    integer i
    sequence reg_list
    if length(oldname) > 3 and length(newname) > 3 then
	if not compare(oldname[2],":") and not compare(newname[2],":") then
	    if compare(oldname[1], newname[1]) then
		i = copy(oldname,newname)
		if not i then
		    return i
		end if
		i = delete(oldname)
		return i
	    end if
	end if
    end if
    low_buff_old = allocate_low(length(oldname) + 1)
    if not low_buff_old then
	return false
    end if
    low_buff_new = allocate_low(length(newname) + 1)
    if not low_buff_new then
	free_low(low_buff_old)
	return false
    end if
    poke(low_buff_old, oldname & 0)
    poke(low_buff_new, newname & 0)
    reg_list = repeat(0,10)
    if dosver() < 7 then
	reg_list[REG_AX] = #5600
    else
	reg_list[REG_AX] = #7156
    end if
    reg_list[REG_DS] = floor(low_buff_old / 16)
    reg_list[REG_DX] = remainder(low_buff_old, 16)
    reg_list[REG_ES] = floor(low_buff_new / 16)
    reg_list[REG_DI] = remainder(low_buff_new, 16)
    reg_list[REG_FLAGS] = or_bits(reg_list[REG_FLAGS], 1)
    reg_list = dos_interrupt(#21, reg_list)
    free_low(low_buff_old)
    free_low(low_buff_new)
    if and_bits(reg_list[REG_FLAGS], 1) != 0 then
	return false
    else
	return true
    end if
end function

global function move(string oldpath, string newpath)
    return rename(oldpath, newpath)
end function

global function chmod(string filename, integer attrib)
    atom low_buff
    sequence reg_list
    low_buff = allocate_low(length(filename) + 1)
    if not low_buff then
	return false
    end if
    poke(low_buff, filename & 0)
    reg_list = repeat(0,10)
    if dosver() < 7 then
	reg_list[REG_AX] = #4301
    else
	reg_list[REG_AX] = #7143
	reg_list[REG_BX] = #0001
    end if
    reg_list[REG_CX] = attrib
    reg_list[REG_DS] = floor(low_buff / 16)
    reg_list[REG_DX] = remainder(low_buff, 16)
    reg_list[REG_FLAGS] = or_bits(reg_list[REG_FLAGS], 1)
    reg_list = dos_interrupt(#21, reg_list)
    free_low(low_buff)
    if and_bits(reg_list[REG_FLAGS], 1) != 0 then
	return false
    else
	return true
    end if
end function

global function deltree(string dirname)
    object info
    integer i
    if not compare(dirname[length(dirname)], "\\") then
	dirname = dirname[1..length(dirname)-1]
    end if
    info = dir(dirname)
    if atom(info) then
	return false
    end if
    if length(info) < 3 then
	return rmdir(dirname)
    end if
    info = info[3..length(info)]
    for loop = 1 to length(info) do
	if not match(info[loop][D_ATTRIBUTES],"d") then
	    i = delete(dirname & "\\" & info[loop][D_NAME])
	    if not i then
		return false
	    end if
	else
	    i = deltree(dirname & "\\" & info[loop][D_NAME])
	    if not i then
		return false
	    end if
	    i = rmdir(dirname & "\\" & info[loop][D_NAME])
	end if
    end for
    return rmdir(dirname)
end function

