/* mod2zzt.c - Converts .mod music to the equivalent ZZT-OOP #play commands
 * Copyright (C) 2007 Chris Mounce. */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "mod2zzt.h"

/*
 *	Main program block
 */
 
modInfo currentMod;

int main (int argc, char* argv[]) {
	FILE* input = NULL;
	FILE* output = NULL;
	int i;
	
	/* arguments */
	char expectedArgs = 2;
	char primaryChannel = -1;
	char secondaryChannel = -1;
	char speed = NORMAL;
	int outputFileName = -1;

	/* Parse arguments. If we get confused, quit with a usage message */
	for (i = 1; i < argc; i++)
		if (strcmp(argv[i], "--help") == 0) printUsage();
	for (i = 1; i < argc; i++) {
		if (strcmp(argv[i], "--slow") == 0) {
			if (speed == FAST) printUsage();
			speed = SLOW;
			expectedArgs++;
		} else if (strcmp(argv[i], "--fast") == 0) {
			if (speed == SLOW) printUsage();
			speed = FAST;
			expectedArgs++;
		} else if (strcmp(argv[i], "-c") == 0) {
			if (i + 1 == argc) printUsage();
			if (primaryChannel < 0) {
				primaryChannel = (char) atoi(argv[i + 1]);
				if (primaryChannel < 1) printUsage();
			} else if (secondaryChannel < 0) {
				secondaryChannel = (char) atoi(argv[i + 1]);
				if (secondaryChannel < 1) printUsage();
			} else printUsage();
			expectedArgs += 2;
			i++;
		} else if (strcmp(argv[i], "-o") == 0) {
			if (i + 1 == argc) printUsage();
			if (output != NULL) printUsage();
			outputFileName = i + 1;
			expectedArgs += 2;
			i++;
		} else {
			if (i != 1) printUsage();
		}
	}
	if (argc < expectedArgs) printUsage();
	if (primaryChannel < 1) primaryChannel = 1;
	if (primaryChannel == secondaryChannel) printUsage();

	/* Whew! Now, attempt to open file */
	input = fopen(argv[1], "rb");
	if (input == NULL) error("Could not open file.");
	
	/* Read the file */
	readModInfo(input, &currentMod);
	if (primaryChannel > currentMod.numChannels || secondaryChannel > currentMod.numChannels)
		error("Specified channel does not exist.");
	readModPatterns(input, &currentMod, primaryChannel, secondaryChannel);
	fclose(input);
	
	/* Convert each pattern to ZZT-OOP */
	printf("Converting to ZZT-OOP...\n\n");
	if (outputFileName > 0) {
		output = freopen(argv[outputFileName], "w", stdout);
		if (output == NULL) error("Could not open file for output.");
	}
	printf("#cycle 4");
	if (secondaryChannel < 1) {
		for (i = 0; i < currentMod.patternTableSize; i++) {
			writeNote(0, 0); /* Reset music system */
			convertPattern1(currentMod.patterns[currentMod.patternTable[i]], speed);
		}
	} else {
		for (i = 0; i < currentMod.patternTableSize; i++) {
			writeNote(0, 0); /* Reset music system */
			convertPattern2(currentMod.patterns[currentMod.patternTable[i]], speed);
		}
	}
	printf("\n");
	
	/* Clean up */
	for (i = 0; i < currentMod.numPatterns; i++) free(currentMod.patterns[i]);
	if (output != NULL) fclose(output);
	
	return 0;
}



/* Prints usage information and exits the program. */
void printUsage () {
	printf("mod2zzt - Converts .mod music to the equivalent ZZT-OOP #play commands\n");
	printf("Usage: mod2zzt file.mod [options]\n");
	printf("Options:\n");
	printf("  -c N     Include channel number N in the conversion. No more than 2 channels\n");
	printf("           may be included.\n");
	printf("  --fast   Makes the conversion play twice as fast as it normally would. Notes\n");
	printf("           may be dropped if multiple channels are in the conversion.\n");
	printf("           Cannot be combined with --slow.\n");	
	printf("  --help   Print this page.\n");
	printf("  -o FILE  Write the ZZT-OOP code to the specified file.\n");
	printf("  --slow   Makes the conversion play half as fast as it normally would.\n");
	printf("           Cannot be combined with --fast.\n");	
	exit(0);
}



/*
 *	Mod processing code
 */

/* Reads the basic data from a .mod file and stores it in the specified modInfo object. */
void readModInfo (FILE* input, modInfo* mod) {
	char magic[4];
	int i;
	unsigned char temp[2];
	printf("Reading mod...\n");
	
	/* We'll start off with these assumptions */
	mod->numChannels = 4;
	mod->numSamples = 31;
	
	/* Read and interpret file magic - these tests were taken from libmodplug */
	fseek(input, 20 + 30*31 + 130, SEEK_SET); /* Jump past title, samplemagic, and pattern table */
	if (fread(&magic, 4, 1, input) != 1) error("Unexpected end of file! File may not be a mod file.");

	if ((isMagic(magic, "M.K.")) || (isMagic(magic, "M!K!")) || (isMagic(magic, "M&K!")) || (isMagic(magic, "N.T."))) mod->numChannels = 4; else
	if ((isMagic(magic, "CD81")) || (isMagic(magic, "OKTA"))) mod->numChannels = 8; else
	if ((magic[0]=='F') && (magic[1]=='L') && (magic[2]=='T') && (magic[3]>='4') && (magic[3]<='9')) mod->numChannels = magic[3] - '0'; else
	if ((magic[0]>='2') && (magic[0]<='9') && (magic[1]=='C') && (magic[2]=='H') && (magic[3]=='N')) mod->numChannels = magic[0] - '0'; else
	if ((magic[0]=='1') && (magic[1]>='0') && (magic[1]<='9') && (magic[2]=='C') && (magic[3]=='H')) mod->numChannels = magic[1] - '0' + 10; else
	if ((magic[0]=='2') && (magic[1]>='0') && (magic[1]<='9') && (magic[2]=='C') && (magic[3]=='H')) mod->numChannels = magic[1] - '0' + 20; else
	if ((magic[0]=='3') && (magic[1]>='0') && (magic[1]<='2') && (magic[2]=='C') && (magic[3]=='H')) mod->numChannels = magic[1] - '0' + 30; else
	if ((magic[0]=='T') && (magic[1]=='D') && (magic[2]=='Z') && (magic[3]>='4') && (magic[3]<='9')) mod->numChannels = magic[3] - '0'; else
	if (isMagic(magic,"16CN")) mod->numChannels = 16; else
	if (isMagic(magic,"32CN")) mod->numChannels = 32; else mod->numSamples = 15;
	
	printf("%d channels detected.\n", mod->numChannels);
	
	/* Read sample volumes */
	printf("Reading sample volumes...\n");
	fseek(input, 20, SEEK_SET);
	for (i = 1; i < mod->numSamples + 1; i++) {
		fseek(input, 22, SEEK_CUR);
		if (fread(temp, 2, 1, input) != 1) error("Unexpected end of file! File may not be a mod file.");
		if (temp[0]*256 + temp[1] < 2) { /* Set volume of empty samples to 0 */
			mod->sampleVolumes[i] = 0;
			fseek(input, 6, SEEK_CUR);
			continue;
		}
		fseek(input, 1, SEEK_CUR);
		if (fread(&(mod->sampleVolumes[i]), 1, 1, input) != 1) error("Unexpected end of file! File may not be a mod file.");
		fseek(input, 4, SEEK_CUR);
	}
	
	/* Read pattern information */
	printf("Reading pattern information...\n");
	if (fread(&(mod->patternTableSize), 1, 1, input) != 1) error("Unexpected end of file! File may not be a mod file.");
	fseek(input, 1, SEEK_CUR);
	if (fread((mod->patternTable), 128, 1, input) != 1) error("Unexpected end of file! File may not be a mod file.");
	mod->numPatterns = 1;
	for (i = 0; i < 128; i++) { /* Find the highest pattern - this tells us the number of patterns */
		if (mod->patternTable[i] + 1 > mod->numPatterns) mod->numPatterns = mod->patternTable[i] + 1;
	}
	if (mod->numSamples != 15) fseek(input, 4, SEEK_CUR);
}



/* Read patterns from the specified channels in a .mod file, and store the
 * results in a modInfo object. To not use a secondary channel, pass a negative
 * value.
 * Assumes that the input file has been fast-forwarded to the point where the
 * pattern data begins. */
void readModPatterns (FILE* input, modInfo* mod, char primaryChannel, char secondaryChannel) {
	/* Loop variables */
	int i, curPattern, curDivision, curChannel;

	/* Variables for reading notes */
	unsigned char temp[4];
	int period;
	unsigned char sample;
	char note;
	
	printf("Reading patterns...\n");

	/* Step through each pattern, division of each pattern, and channel of each division */
	for (curPattern = 0; curPattern < mod->numPatterns; curPattern++) {
		mod->patterns[curPattern] = malloc(sizeof(pattern));
		for (curDivision = 0; curDivision < 64; curDivision++) {
			for (curChannel = 1; curChannel < mod->numChannels + 1; curChannel++) {
				/* If we don't want this channel, ignore it. */
				if (curChannel != primaryChannel && curChannel != secondaryChannel) {
					fseek(input, 4, SEEK_CUR);
					continue;
				}
				
				/* Read the period and sample */
				if (fread(temp, 4, 1, input) != 1) error("Unexpected end of file! File may not be a mod file.");
				period = (temp[0] & 0x0F)*256 + temp[1];
				sample = (temp[0] & 0xF0) + (temp[2] & 0xF0)/16;
				
				/* Convert it all to a note */
				if (mod->sampleVolumes[sample] == 0) {
					note = 0;
				} else if (period == 0) {
					note = -1;
				} else {
					note = 0;
					while (period <= NOTE_MIDPOINTS[note]) note++;
					note++;
				}
				
				/* Store the note */
				if (curChannel == primaryChannel) (mod->patterns[curPattern])->primary[curDivision] = note;
				else (mod->patterns[curPattern])->secondary[curDivision] = note;
			}
		}
	}
}



/* Compares the first 4 bytes of s1 and s2. Returns 0 if they match,
	nonzero otherwise */
int isMagic (char* s1, char* s2) {
	return !strncmp(s1, s2, 4);
}



/*
 * Conversion code
 */

/* Converts a one-channel pattern to ZZT-OOP and prints it to stdout */
void convertPattern1 (pattern* p, char speed) {
	ZZTmusic music;
	int i, j;
	
	/* Step through each division, turning each into notes */
	j = 0;
	for (i = 0; i < 64; i++) {
		music.notes[j++] = p->primary[i];
		if (speed != FAST) music.notes[j++] = -1;
		if (speed == SLOW) {
			music.notes[j++] = -1;
			music.notes[j++] = -1;
		}
	}
	music.size = j;
	
	/* Convert the notes to ZZT-OOP */
	convertPattern(&music);
}



/* Converts a two-channel pattern to ZZT-OOP and prints it to stdout */
void convertPattern2 (pattern* p, char speed) {
	ZZTmusic music;
	int i;
	int j = 0;
	
	/* Break notes into the smallest notes possible (1/32 notes) */
	if (p->primary[0] < 0) p->primary[0] = 0;
	if (p->secondary[0] < 0) p->secondary[0] = 0;
	for (i = 1; i < 64; i++) {
		if (p->primary[i] < 0) p->primary[i] = p->primary[i - 1];
		if (p->secondary[i] < 0) p->secondary[i] = p->secondary[i - 1];
	}
	
	/* Step through each division, turning each into notes */
	for (i = 0; i < 64; i++) {
		switch (speed) {
		case FAST:
			music.notes[j++] = p->primary[i];
			music.notes[j++] = p->secondary[++i];
			break;
		case SLOW:
			music.notes[j++] = p->primary[i];
			music.notes[j++] = p->secondary[i];
			/* Fall through */
		case NORMAL:
			music.notes[j++] = p->primary[i];
			music.notes[j++] = p->secondary[i];
		}
	}
	music.size = j;
	
	/* Convert the notes to ZZT-OOP */
	convertPattern(&music);
}



/* Prints the ZZT-OOP code that plays the specified music */
void convertPattern (ZZTmusic* music) {
	/* Loop variables */
	unsigned int position;
	int i = 0;
	
	/* Musical stuff */
	unsigned int curNoteLength = 0;
	char curNote = 0;
	
	/* Join rests together if it makes storage more efficient */
	if (music->notes[0] < 0) music->notes[0] = 0;
	while (i < music->size) {
		/* Find next rest */
		while (i < music->size) {
			if (music->notes[i] == 0) break;
			i++;
		}
		i++;
		/* Is it worth it to join these rests? */
		if (i + 2 < music->size && music->notes[i] <= 0 && music->notes[i + 1] <= 0 && music->notes[i + 2] <= 0) {
			/* Yes. Turn all following rests to -1 */
			while (i < music->size) {
				if (music->notes[i] > 0) break;
				music->notes[i] = -1;
				i++;
			}
		}
	}
	
	/* Step through each note */
	curNote = music->notes[0];
	curNoteLength = 1;
	position = 1;
	while (position < music->size) {
		if (music->notes[position] == -1) {
			curNoteLength++; /* Add on to the current note */
		} else {
			/* Write out the current note */
			while (curNoteLength > 0) {
				i = 0;
				while (EASY_DURATIONS[i] > curNoteLength) i++;
				writeNote(curNote, EASY_DURATIONS[i]);
				curNoteLength -= EASY_DURATIONS[i];
			}
			
			/* Read the next note */
			curNote = music->notes[position];
			curNoteLength = 1;
		}
		position++;
	}
	
	/* Write out that final note */
	while (curNoteLength > 0) {
		i = 0;
		while (EASY_DURATIONS[i] > curNoteLength) i++;
		writeNote(curNote, EASY_DURATIONS[i]);
		curNoteLength -= EASY_DURATIONS[i];
	}
}



/* Writes a note to stdout, accounting for end-of-line situations and the like.
 * To reset the system, call with a duration of 0.
 * Assumes that the requested duration is convenient. */
void writeNote (char note, unsigned char duration) {
	static char lastOctave = 0;
	static char lastDuration = 1;
	static char lineLength = 42;
	static unsigned int durationOfLine = 0;
	
	char octave;
	char noteLength;

	int i;
	
	/* Estimate the length of the note we're going to write */
	if (note == 0) noteLength = 1;
	else noteLength = strlen(NOTE_CODES[(note - 1) % 12]);
	if (duration != lastDuration) {
		switch (duration) {
		case 108:
		case 54:
		case 27:
			noteLength += 4;
			break;
		case 72:
		case 36:
		case 18:
		case 9:
			noteLength += 3;
			break;
		case 48:
		case 24:
		case 12:
		case 6:
		case 3:
			noteLength += 2;
			break;
		default:
			noteLength += 1;
		}
	}
	if (note != 0) {
		octave = (note - 1)/12 - 2;
		noteLength += abs(lastOctave - octave);
	} else {
		octave = lastOctave;
	}
	
	/* Break onto a new line if we're (1) out of space or (2) ordered to */
	if (lineLength + noteLength > 42 || duration == 0) {
		/* Reset music stats */
		lastOctave = 0;
		lastDuration = 1;

		/* Print idles */
		putchar('\n');
		lineLength = 0;
		while (durationOfLine >= 8) {
			durationOfLine -= 8;
			printf("/i");
			lineLength += 2;
			if (lineLength == 42) {
				putchar('\n');
				lineLength = 0;
			}
		}
		
		
		/* Print the beginning of the #play statement */
		if (lineLength != 0) putchar('\n');
		printf("#play ");
		lineLength = 6;

		/* If we're just resetting the system, eject here */
		if (duration == 0) return;
	}
	
	/* Print octave, if we need it */
	switch (octave - lastOctave) {
	case -2:
		putchar('-');
		lineLength++;
	case -1:
		putchar('-');
		lineLength++;
		break;
	case 2:
		putchar('+');
		lineLength++;
	case 1:
		putchar('+');
		lineLength++;
	default:
		/* Do nothing */;
	}
	
	/* Print duration if necessary */
	if (duration != lastDuration) {
		i = 0;
		while (duration != EASY_DURATIONS[i]) i++;
		printf(DURATION_CODES[i]);
		lineLength += strlen(DURATION_CODES[i]);
	}
	
	/* And we always need to print the note. */
	if (note == 0) {
		putchar('x');
		lineLength++;
	} else {
		printf(NOTE_CODES[(note - 1) % 12]);
		lineLength += strlen(NOTE_CODES[(note - 1) % 12]);
	}
	
	/* Update info */
	lastOctave = octave;
	lastDuration = duration;
	durationOfLine += duration;
}



/*
 *	Miscellaneous
 */

/* Displays the specified error message and halts the program*/
void error (char* message) {
	int i;
	printf("ERROR: %s\n", message);
	for (i = 0; i < currentMod.numPatterns; i++) free(currentMod.patterns[i]);
	exit(1);
}