// $Id$
// ZZT++
//
// Copyright (C) 1998 Wouter E.P. Bovelander
//	wbovela@wxs.nl http://welcome.to/wep
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License as
// published by the Free Software Foundation; either version 2 of
// the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//

#include <stdio.h>
#include <malloc.h>
#include <stdlib.h>
#include <mem.h>
#include <string.h>
#include <ctype.h>

// ==============================================================
//
//	Data section
//
// ==============================================================

// this struct is the same for ZZT and SAV files
typedef struct GameInfo
{
	int dummy1; // FFFF
	int NoBoards;
	int NoAmmo;
	int NoGems;
	unsigned char BlueKey;
	unsigned char GreenKey;
	unsigned char CyanKey;
	unsigned char RedKey;
	unsigned char PurpleKey;
	unsigned char YellowKey;
	unsigned char WhiteKey;
	int Health;
	int dummy2;
	int NoTorches;
	int NoTorchCycles;
	int NoEnergizerCycles;
	int dummy3; // 0000
	int Score;
	unsigned char TitleLength;
	char Title[20]; // 20 positions
	unsigned char Flag1Length;
	char Flag1Name[20]; // 20 positions
	unsigned char Flag2Length;
	char Flag2Name[20]; // 20 positions
	unsigned char Flag3Length;
	char Flag3Name[20]; // 20 positions
	unsigned char Flag4Length;
	char Flag4Name[20]; // 20 positions
	unsigned char Flag5Length;
	char Flag5Name[20]; // 20 positions
	unsigned char Flag6Length;
	char Flag6Name[20]; // 20 positions
	unsigned char Flag7Length;
	char Flag7Name[20]; // 20 positions
	unsigned char Flag8Length;
	char Flag8Name[20]; // 20 positions
	unsigned char Flag9Length;
	char Flag9Name[20]; // 20 positions
};

// Saved game indicator at position 264
unsigned char GameSaved;

struct Board
{
	unsigned int BoardSize;
	unsigned char TitleLength;
	char Title[34]; // Max length is 34 bytes
	unsigned char dummy[16]; // 16 bytes of mysterious functionality
	// Usually 06 0e 00 90 00 00 AF 4C 04 0F 02 12 B4 00 00 55

	// Blocks follow here. The number of blocks is to be determined
	// dynamically (i.e. until sum(Block.NoChars)==1500)
};

// Screens are compressed by replacing blocks of repeating characters by
// structs like these. Single characters are also stored this way. Since
// boards usually contain repeating consecutive sections, this way of
// storing screens will result in saving memory space.
struct Block
{
	// A screen consists of a maximum of 1500 (0x5dc) characters.
	// The sum of NoChars should be 1500.
	unsigned char NoChars;
	unsigned char CharCode;
	unsigned char ColCode;
};

struct BoardInfo
{
	unsigned char MaxShots;
	unsigned char DarkBoard;
	unsigned char BoardNorth;
	unsigned char BoardSouth;
	unsigned char BoardWest;
	unsigned char BoardEast;
	unsigned char ReEnter;
	unsigned char MessageLength;
	char Message[58];
	int dummy1; // 0x71b
	unsigned int TimeLimit;
	char dummy2[16]; // zeroes
	unsigned int NoObjects;

	// Board objects follow here. During runtime each object will be
	// represented by an instance of (a subclass of) the class Object.
};

struct BoardObject
{
	unsigned char XPos;
	unsigned char YPos;
	int XMovementStep;
	int YMovementStep;
	unsigned int CycleSpeed;
	unsigned char Magic1; // Too much functionality to fit into symbol
	unsigned char Magic2; // description.
	unsigned char DestinationBoardNo; // Ony for passages, otherwise 00
	int dummy1; // always FF FF
	int dummy2; // always FF FF
	unsigned char CharCodeUnder;
	unsigned char ColCodeUnder;
	char *ProgramTextPtr; // Pointer to programs in memory ???
	unsigned int ProgramIP; // Instruction Pointer
	// 10 bytes from here!
	unsigned int ProgramLength;
	int d1, d2, d3, d4;
};

// ==============================================================
//
//	Global variables (private members in C++)
//
// ==============================================================

// These are global to this test program but should be
// private members to BM

// One struct for GameInfo
GameInfo GI;

// An array of Boards
// Example: BRD[0].BoardSize
Board *BRD;

// An array of blocks per board.
// Example: BLCK[0][0].NoChars is block# 0 of board# 0
Block **BLCK;

// An array of BoardInfo structs
// Example: BRDINF[0] is the info for board 0.
BoardInfo *BRDINF;

// An array of objects per board.
// Example: BRDOBJ[0][1] is the second object for board# 0
BoardObject **BRDOBJ;

// An array of Program texts per object.
char ***OBJPRG;

// ==============================================================
//
//	Functions
//
// ==============================================================

void PrintGameInfo(void);
void PrintBoardInfo(int boardcounter);
void PrintBoard(int boardcounter);
void PrintObjectInfo(int boardcounter, int objectcounter);

int ReadWorld(char *file, int screen)
{
	FILE *world;
	int board, blocksum, block, object;

	// Open world
	if ((world = fopen(file, "rb"))== NULL)
	{
		printf("File %s was not found.\n", file);
		return 1;
	}
	fprintf(stdout,"START OF TRACE.\n===============\n");
	fprintf(stdout,"Opened file %s.\n\n",file);

	//
	// Read GameInfo struct
	//
	fread((void *)&GI, sizeof(GameInfo), 1, world);
	PrintGameInfo();

	// Array of Board structs.
	if ((BRD=(Board *)malloc(GI.NoBoards*sizeof(Board)))==NULL)
		return 10;

	// Array of arrays of Block structs.
	if ((BLCK=(Block **)malloc(GI.NoBoards*sizeof(Block *)))==NULL)
		return 11;

	// Array of BoardInfo structs.
	if ((BRDINF=(BoardInfo *)malloc(GI.NoBoards*sizeof(BoardInfo)))==NULL)
		return 12;

	// Array of BoardObject structs.
	if ((BRDOBJ=(BoardObject **)malloc(GI.NoBoards*sizeof(BoardObject *)))==NULL)
		return 13;

	// Array of Object Program strings.
	if ((OBJPRG=(char ***)malloc(GI.NoBoards*sizeof(char **)))==NULL)
		return 14;

	//
	// Seek to Game Saved indicator
	//
	fseek(world, 264, SEEK_SET);
	fread((void *)&GameSaved, sizeof(unsigned char), 1, world);
	fprintf(stdout,"GameSaved=%d\n\n",GameSaved);

	//
	// Seek to Start of Boards
	//
	fseek(world, 512, SEEK_SET);

	//	Start reading boards
	// There is always one title screen board. So use <= in stead of <
	for (board=0; board<=GI.NoBoards; board++)
	{
		//
		// Read the Board struct into BRD array
		//
		fread(&(BRD[board]), sizeof(Board), 1, world);
		PrintBoard(board);

		if (screen)
			fprintf(stdout,"Compressed screen information.\n===========================\n");
		// realloc() requires this pointer to be NULL for the first call
		BLCK[board]=NULL;
		blocksum=block=0;
		while (blocksum<1500)
		{
			// Request more memory for this block
			if ((BLCK[board]= \
				(Block *)realloc(BLCK[board], (block+1)*sizeof(Block)))==NULL)
					return 15;

			//
			// Read the block into memory.
			//
			fread(&(BLCK[board][block]), sizeof(Block), 1, world);

			// Sum up the number of characters.
			blocksum+=BLCK[board][block].NoChars;

			if (screen)
				fprintf(stdout,"\tBlock#=%d, NoChars=%d, CharCode=%d, ColCode=%d, Sum=%d\n", \
			block,BLCK[board][block].NoChars, \
			BLCK[board][block].CharCode, \
			BLCK[board][block].ColCode, \
			blocksum);

			block++;
		}
		if (screen)
			fprintf(stdout,"End Compressed screen information.\n===============================\n\n");

		//
		// Read BoardInfo struct
		//
		fread(&(BRDINF[board]), sizeof(BoardInfo), 1, world);
		PrintBoardInfo(board);

		// realloc() requires these pointers to be NULL for the first call
		BRDOBJ[board]=NULL;
		OBJPRG[board]=NULL;

		object=0;
		// There is always one player object. So use <= in stead of <
		while (object<=BRDINF[board].NoObjects)
		{
			// Request more memory for this object
			if ((BRDOBJ[board]= \
				(BoardObject *)realloc(BRDOBJ[board], (object+1)*sizeof(BoardObject)))==NULL)
					return 16;

			//
			// Read the object into memory.
			//
			fread(&(BRDOBJ[board][object]), sizeof(BoardObject), 1, world);
			PrintObjectInfo(board, object);

			// request memory for a pointer to program text
			// even though there might not be a program.
			if ((OBJPRG[board]= \
				(char **)realloc(OBJPRG[board], (object+1)*sizeof(char *)))==NULL)
					return 17;

			// Only if there is a program to be read.
			// Never read program of player object.
			if ((BRDOBJ[board][object].ProgramLength>0) && object>0)
			{
				// allocate memory for this text. Use length from object struct.
				if ((OBJPRG[board][object]= \
					(char *)malloc(BRDOBJ[board][object].ProgramLength))==NULL)
						return 18;

				//
				// Read program text into memory.
				//
				fread(OBJPRG[board][object], \
					BRDOBJ[board][object].ProgramLength, 1, world);
				fprintf(stdout, "Program=\n%s\n",OBJPRG[board][object]);
			}
			// Count this object
			object++;
		} // End of Object loop
	} // End of Board loop

	// Cose file
	fflush(world);
	fclose(world);
	return 0;
}

//
// Print Object Information
//
void PrintObjectInfo(int board, int object)
{
	fprintf(stdout,"\nObject %d of %d %s\n================\n", \
		object, BRDINF[board].NoObjects, object==0?"(player)\0":"\0");
	fprintf(stdout,"XPos:                  %d\n",BRDOBJ[board][object].XPos);
	fprintf(stdout,"YPos:                  %d\n",BRDOBJ[board][object].YPos);
	fprintf(stdout,"XMovementStep:         %d\n",BRDOBJ[board][object].XMovementStep);
	fprintf(stdout,"YMovementStep:         %d\n",BRDOBJ[board][object].YMovementStep);
	fprintf(stdout,"CycleSpeed:            %d\n",BRDOBJ[board][object].CycleSpeed);
	fprintf(stdout,"Magic1:                %d\n",BRDOBJ[board][object].Magic1);
	fprintf(stdout,"Magic2:                %d\n",BRDOBJ[board][object].Magic2);
	fprintf(stdout,"DestinationBoardNo:    %d\n", BRDOBJ[board][object].DestinationBoardNo);
	fprintf(stdout,"Dummy1:                %u\n",BRDOBJ[board][object].dummy1);
	fprintf(stdout,"Dummy2:                %u\n",BRDOBJ[board][object].dummy2);
	fprintf(stdout,"CharCodeUnder:         %d\n",BRDOBJ[board][object].CharCodeUnder);
	fprintf(stdout,"ColCodeUnder:          %d\n",BRDOBJ[board][object].ColCodeUnder);
	fprintf(stdout,"ProgramTextPtr:        %p\n",BRDOBJ[board][object].ProgramTextPtr);
	fprintf(stdout,"ProgramIP:             %d\n",BRDOBJ[board][object].ProgramIP);
	fprintf(stdout,"ProgramLength:         %d\n",BRDOBJ[board][object].ProgramLength);
	fprintf(stdout,"End Object Info\n===============\n");
}

//
// Print Board struct
//
void PrintBoard(int board)
{
	int i;

	fprintf(stdout,"\nBoard Header %d of %d\n====================\n", \
		board, GI.NoBoards);
	fprintf(stdout,"BoardSize:       %d\n",BRD[board].BoardSize);
	fprintf(stdout,"Titlelength:     %d\n",BRD[board].TitleLength);
	fprintf(stdout,"Title:           %.*s\nDummy=(hex) ", BRD[board].TitleLength, \
		BRD[board].Title);

	for (i=0;i<16;i++)
		fprintf(stdout,"%02x ",BRD[board].dummy[i]);

	fprintf(stdout,"\nEnd Board Header\n================\n\n");
}

//
// Print GI
//

void PrintGameInfo(void)
{
	fprintf(stdout,"Game Info\n=========\n");
	fprintf(stdout,"dummy1:         %u\n",GI.dummy1);
	fprintf(stdout,"NoBoards:       %d\n",GI.NoBoards);
	fprintf(stdout,"NoAmmo:         %d\n",GI.NoAmmo);
	fprintf(stdout,"NoGems:         %d\n",GI.NoGems);
	fprintf(stdout,"BlueKey:        %d\n",GI.BlueKey);
	fprintf(stdout,"GreenKey:       %d\n",GI.GreenKey);
	fprintf(stdout,"CyanKey:        %d\n",GI.CyanKey);
	fprintf(stdout,"RedKey:         %d\n",GI.RedKey);
	fprintf(stdout,"PurpleKey:      %d\n",GI.PurpleKey);
	fprintf(stdout,"YellowKey:      %d\n",GI.YellowKey);
	fprintf(stdout,"WhiteKey:       %d\n",GI.WhiteKey);
	fprintf(stdout,"Health:         %d\n",GI.Health);
	fprintf(stdout,"dummy2:         %u\n",GI.dummy2);
	fprintf(stdout,"NoTorches:      %d\n",GI.NoTorches);
	fprintf(stdout,"NoTorchCycles:  %d\n",GI.NoTorchCycles);
	fprintf(stdout,"NoEnergizerCycles:%d\n",GI.NoEnergizerCycles);
	fprintf(stdout,"dummy3:         %u\n",GI.dummy3);
	fprintf(stdout,"Score:          %d\n",GI.Score);
	fprintf(stdout,"TitleLength:    %d\n",GI.TitleLength);
	fprintf(stdout,"Title:          %.*s\n",GI.TitleLength, GI.Title);
	fprintf(stdout,"Flag1Length     %d\n",GI.Flag1Length);
	fprintf(stdout,"Flag1Name       %.*s\n",GI.Flag1Length, GI.Flag1Name);
	fprintf(stdout,"Flag2Length     %d\n",GI.Flag2Length);
	fprintf(stdout,"Flag2Name       %.*s\n",GI.Flag2Length, GI.Flag2Name);
	fprintf(stdout,"Flag3Length     %d\n",GI.Flag3Length);
	fprintf(stdout,"Flag3Name       %.*s\n",GI.Flag3Length, GI.Flag3Name);
	fprintf(stdout,"Flag4Length     %d\n",GI.Flag4Length);
	fprintf(stdout,"Flag4Name       %.*s\n",GI.Flag4Length, GI.Flag4Name);
	fprintf(stdout,"Flag5Length     %d\n",GI.Flag5Length);
	fprintf(stdout,"Flag5Name       %.*s\n",GI.Flag5Length, GI.Flag5Name);
	fprintf(stdout,"Flag6Length     %d\n",GI.Flag6Length);
	fprintf(stdout,"Flag6Name       %.*s\n",GI.Flag6Length, GI.Flag6Name);
	fprintf(stdout,"Flag7Length     %d\n",GI.Flag7Length);
	fprintf(stdout,"Flag7Name       %.*s\n",GI.Flag7Length, GI.Flag7Name);
	fprintf(stdout,"Flag8Length     %d\n",GI.Flag8Length);
	fprintf(stdout,"Flag8Name       %.*s\n",GI.Flag8Length, GI.Flag8Name);
	fprintf(stdout,"Flag9Length     %d\n",GI.Flag9Length);
	fprintf(stdout,"Flag9Name       %.*s\n",GI.Flag9Length, GI.Flag9Name);
	fprintf(stdout,"\nEnd Game Info\n=============\n");
};

//
// Print BoardInfo
//
void PrintBoardInfo(int board)
{
	int i;

	fprintf(stdout,"\nBoard Footer %d of %d\n====================\n", \
		board, GI.NoBoards);

	fprintf(stdout,"MaxNoShots:     %d\n",BRDINF[board].MaxShots);
	fprintf(stdout,"DarkBoard:      %d\n",BRDINF[board].DarkBoard);
	fprintf(stdout,"BoardNorth:     %d\n",BRDINF[board].BoardNorth);
	fprintf(stdout,"BoardSouth:     %d\n",BRDINF[board].BoardSouth);
	fprintf(stdout,"BoardWest:      %d\n",BRDINF[board].BoardWest);
	fprintf(stdout,"BoardEast:      %d\n",BRDINF[board].BoardEast);
	fprintf(stdout,"ReEnter:        %d\n",BRDINF[board].ReEnter);
	fprintf(stdout,"MessageLength:  %d\n",BRDINF[board].MessageLength);
	fprintf(stdout,"Message:        %.*s\n",BRDINF[board].MessageLength,\
		BRDINF[board].Message);
	fprintf(stdout,"dummy1:         %u\n",BRDINF[board].dummy1);
	fprintf(stdout,"TimeLimit:      %d\n",BRDINF[board].TimeLimit);

	fprintf(stdout,"dummy2:(hex) ");
	for (i=0;i<16;i++)
		fprintf(stdout,"%02x ",BRDINF[board].dummy2[i]);

	fprintf(stdout,"\nNoObjects:    %d\n",BRDINF[board].NoObjects);
	fprintf(stdout,"End Board Footer\n================\n\n");
}


//
// Function: ParseCmdLine
//
void ParseCmdLine(int argc, char **argv, char *world, int *Screen)
{
	char *walk;
	int i;

	// first parameter is always program name, so start with 1
	for (i=1;i<argc;i++)
	{
		// Convert parameter to upper case.
		walk=argv[i];
		while (*walk=toupper(*walk)) walk++;

		// Copy filename into world.
		if (i==1)
		{
			strcpy(world, argv[i]);

			// Look for file extention
			// If found then don't add .ZZT
			walk=world;
			while (*walk!='.' && *walk) walk++;
			if (*walk=='.') return;

			// No file extention so add .ZZT
			if (strncmp(argv[i]+strlen(argv[i])-4,".ZZT",4))
			{
				 world=strcat(world,".ZZT\0");
			}
		}

		// Second parameter - means no screen information
		if (i==2)
		{
			if (!strcmp(argv[i],"-"))
				*Screen=0;
		}
	}
}

//
// Main
//
int main(int argc, char **argv, char **env)
{
	char *world;
	int err=0, screen=1;

	// Parse cmd line
	world=(char *)malloc(13*sizeof(char));
	memcpy(world,"TOWN.ZZT\0",10);
	ParseCmdLine(argc, argv, world, &screen);

	// Read GameInfo struct
	if (err=ReadWorld(world, screen))
	{
		fprintf(stdout, "Error: %d\n", err);
	}

	return err;
}
